#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <termios.h>
#include <unistd.h>
#include <sys/resource.h>
#include <dirent.h>
#include <sys/stat.h>
#include <pthread.h>
#include "mpg123.h"
#include "mpglib.h"
#include <linux/soundcard.h>
#include "tag.h"

#define AUDIO_DEV "/dev/dsp"
#define AUDIO_BUF 0x4000
#define MAXFILES  0x4000
#define MAXFLEN   (32<<20)

// Set canonical mode for stdin (line editing)
void setcanon()
{
        struct termios term;
        tcgetattr(0,&term);
        term.c_lflag|=ICANON;
        term.c_lflag|=ECHO;
        if (tcsetattr(0,TCSANOW,&term)) {
                fprintf(stderr,"tcsetattr failed\n");
                exit(0);
        }
}
// Set NOcanonical mode for stdin (single character input)
// gcc-2.95.3(i386) generates buggy code for this function
// (uncontrolled stack pointer)

void setnocanon()
{
        struct termios term;
        tcgetattr(0,&term);
        term.c_lflag&=~ICANON;
        term.c_lflag&=~ECHO;
        // Solaris is a bit stupid fetching 4 chars at a time by default.
        // It tries to be "more efficient". But I think noncanonical mode
        // is used for reading 1 char at a time 99.9(period) % of times
        term.c_cc[VMIN]=1;      // Force 1 char at a time (too smart Sunos)
        term.c_cc[VTIME]=0;
        if (tcsetattr(0,TCSANOW,&term)) {
                fprintf(stderr,"tcsetattr failed\n");
                exit(0);
        }
}


int audio_fd=0,audio_rate=0,audio_stereo=0,audio_bits=16;

void audio_8bits(short *buf,int size)
{
        int i,n;
        unsigned int j;
        static unsigned int er1,er2,er3,el1,el2,el3;
        short *ps;
        unsigned char *pd;

        n=size/2; if (audio_stereo) n/=2;

        ps=buf; pd=(unsigned char *)buf;
        for (i=0;i<n;i++){
            j=(*ps++)+32768;
            j+=3*el1-3*el2+el3;
            el3=el2; el2=el1;
            el1=j&0x00ff;
            *pd++=j>>8;

            if (audio_stereo) {
               j=(*ps++)+32768;
               j+=3*er1-3*er2+er3;
               er3=er2; er2=er1;
               er1=j&0x00ff;
               *pd++=j>>8;
            }
        }
        write(audio_fd,buf,size/2);
}

void init_audio()
{
	if ((audio_fd=open(AUDIO_DEV, O_WRONLY)) == -1) {
		fprintf(stderr, "Cannot open sound device");
		exit(0);
	}

        if (ioctl(audio_fd, SNDCTL_DSP_SAMPLESIZE, &audio_bits) == EOF) {
                fprintf(stderr, "Cannot set %d bits/sample\n",audio_bits);
        }
        if (ioctl(audio_fd, SNDCTL_DSP_STEREO, &audio_stereo) == EOF) {
                fprintf(stderr, "Cannot set %ld channel(s)\n",audio_stereo+1);
        }
        if (ioctl(audio_fd, SNDCTL_DSP_SPEED, &audio_rate) == EOF) {
                fprintf(stderr, "Cannot set rate %d Hz\n",audio_rate);
        }
        printf("@@ Audio Settings: %d bits, fs=%d Hz, %s\n",
                audio_bits,audio_rate,(audio_stereo)?"Stereo":"Mono");
}

//------------------------------
// MP3 decoding funtion
// Relies on mpg123 library

struct mpstr mp;

int play_block(unsigned char *buf, int len)
{
        int i,j,k,ib,size,done,nf;
        unsigned char a,out[0x4000];
        struct frame ffr;

        InitMP3(&mp);

        // Check for tag & remove it if present
        if (buf[len-128]=='T' && buf[len-127]=='A' && buf[len-126]=='G') {
            printf("++ TAG found:\n");
            len-=128;
            buf[len+32]=0; printf("\t%s\n",&buf[len+ 3]);
            buf[len+62]=0; printf("\t%s\n",&buf[len+33]);
            buf[len+92]=0; printf("\t%s\n",&buf[len+63]);
            a=buf[len+127];
            if (a<GENRE_LAST) printf("\t(%s)\n",genre_list[a]);
        }

        //check for junk at the begining
        for (ib=0;ib<len-1;ib++) {
            if (buf[ib]==0xff && (buf[ib+1]&0xf0)==0xf0) break;
        }
        if (ib) printf("++ Junk at the beginning. (%d bytes)\n",ib);

        // Get audio information from first header:
        // and reopen audio device if needed

        j=(buf[ib]<<24)|(buf[ib+1]<<16)|(buf[ib+2]<<8)|buf[ib+3];
        decode_header(&ffr,j);
        //printf("framesize = %d\n",ffr.framesize);
        k=freqs[ffr.sampling_frequency];
        j=ffr.stereo-1;

        if (k!=audio_rate || j!=audio_stereo) {
            if (audio_fd) close(audio_fd);
            audio_rate=k; audio_stereo=j;
            init_audio();
        }

        // Decoder loop
        done=0;
        for(i=ib;i<len;) {
                //Decode a single frame
                j=(buf[i]<<24)|(buf[i+1]<<16)|(buf[i+2]<<8)|buf[i+3];
                decode_header(&ffr,j); // get new framesize
                decodeMP3(&mp,&buf[i],ffr.framesize+4,out,AUDIO_BUF,&size);
                if (audio_bits==8) audio_8bits((short *)out,size);
                else write(audio_fd,out,size);
                if (read(0,&a,1)>0) {
                    if (a=='n'||a=='q'||a=='r'||a=='p' ||a=='f' ||a=='b' )
                    {done=a;}
                }
                i+=ffr.framesize+4;
                printf("\r++ %d/%d KB (%d%%) ",i>>10,len>>10,i/(len/100));
                fflush(stdout);
                // Advance ~10 sec
                if (done=='f') {
                    for (k=0;k<400;k++) {
                        j=(buf[i]<<24)|(buf[i+1]<<16)|(buf[i+2]<<8)|buf[i+3];
                        decode_header(&ffr,j);
                        i+=ffr.framesize+4;
                        if (i>len) return 0;
                    }
                    done=0;
                }
                // Rewind ~10sec
                if (done=='b') {
                    k=(i-ib)/(ffr.framesize+4); // approx. current frame
                    k-=400; if (k<0) k=0;
                    i=ib;
                    for (;k;k--) {
                        j=(buf[i]<<24)|(buf[i+1]<<16)|(buf[i+2]<<8)|buf[i+3];
                        decode_header(&ffr,j);
                        i+=ffr.framesize+4;
                        if (i>len) return 0;
                    }
                    done=0;
                }
                if (done) break;
        }

        ExitMP3(&mp);

        return done;
}

//-------------------------------------------
// Recursive function for directory scanning

void dir_add(char *pdir,char **pfn, int *ixf)
{
        DIR *dp;
        struct dirent ep,*pep,**namelist;
        struct stat st;
        char *p,buf[1024];
        int i,n,m;

        if ((dp=opendir(pdir))==NULL) return;
        m = scandir(pdir, &namelist, 0, alphasort);
        if (m<0) return;
        //*** Adding current dir. files
        for (i=0;i<m;i++) {
            sprintf(buf,"%s/%s",pdir,namelist[i]->d_name);
            stat(buf,&st);
            n=strlen(buf);
            if (n<4) continue;
            if (S_ISREG(st.st_mode)) {
                p=&buf[n-4];
                if (strcmp(p,".mp3")==0 || strcmp(p,".MP3")==0) {
                    if ((p=malloc(n+1))==NULL) exit(1);
                    strcpy(p,buf);
                    pfn[(*ixf)++]=p;
                }
            }
        }
        //*** scanning Subdirectories
        for (i=0;i<m;i++) {
            sprintf(buf,"%s/%s",pdir,namelist[i]->d_name);
            stat(buf,&st);
            n=strlen(buf);
            if (n<4) continue;
            if (S_ISDIR(st.st_mode)) {
                if (buf[n-1]=='.') continue;
                dir_add(buf,pfn, ixf);
            }
        }

        free(namelist);
        closedir(dp);
}

//--------------------------------------------
// A thread for concurrent file reading
// When rdfn != NULL the "rdfn" file is read
// After file read, rdfn is set to NULL again
// Memory is allocated here

char *bbuf;
int blen;
char *rdfn;

void thread_read_file()
{
    int fd;

    // Low priority for this thread
    setpriority(PRIO_PROCESS,0,10);
    for (;;) {
        while (rdfn==NULL) usleep(20000);
        blen=0;
        printf("\r## Reading %s\n",rdfn);
        if ((fd=open(rdfn,O_RDONLY))==-1){rdfn=NULL; continue; }
        blen=lseek(fd,0,SEEK_END);
        lseek(fd,0,SEEK_SET);
        bbuf=malloc(blen);
        mlock (bbuf,blen); // Avoid swapping
        blen=read(fd,bbuf,blen);
        close(fd);
        rdfn=NULL;
    }
}

#define MAXPL 16

main(int argc, char **argv)
{
        int i,j,n,npl,nd,nf,ixf,random=0;
        char a,buffer[1024],*p,*df,*pdir[MAXPL],*pln[MAXPL],*pfn[MAXFILES];
        FILE *fp;
        uid_t euid, ruid;
        int egid, rgid;
        pthread_t pth;

        // User & Group ID stuff
        ruid = getuid ();
        euid = geteuid ();
        rgid = getgid ();
        egid = getegid ();
        // Do not use root privileges while parsing options
        if (seteuid (ruid)<0) exit(1);
        if (setegid (rgid)<0) exit(1);

        // parse options
        if (argc<2) {
                printf("\n\nMP3PLAY %s\n\n",VERSION);
                printf("Usage: %s [options] files\n",argv[0]);
                printf("Options:\n");
                printf("-z        : Random Play\n");
                printf("-@ <file> : Add MP3 files from <file>\n");
                printf("-s <dir>  : Scan <dir> tree and add MP3 files\n");
                printf("-d <file> : Dump playlist to <file>\n");
                printf("-8        : 8 bit output (with dittering)\n");
                printf("Up to %d playlist files and %d directories can be stated on command line\n\n",MAXPL,MAXPL);
                printf("stdin (Keyboard) controls:\n");
                printf("<f> : Forward advance (10 sec)\n");
                printf("<b> : Backward rewind (10 sec)\n");
                printf("<n> : Next file\n");
                printf("<r> : Repeat same file\n");
                printf("<p> : Previous file (implies pipeline flush)\n");
                printf("<q> : Quit\n");
                exit(0);
        }
        npl=nd=0; df=NULL;
        for (i=1;i<argc;i++) {
            if (*argv[i]=='-') {
                switch(argv[i][1]) {
                   case 'z' : random=1; break;
                   case '@' : if (npl<MAXPL-1) pln[npl++]=argv[i+1];
                              i++; break;
                   case 's' : if (nd<MAXPL-1) pdir[nd++]=argv[i+1];
                              i++; break;
                   case 'd' : df=argv[i+1]; i++; break;
                   case '8' : audio_bits=8; break;
                }
            } else pfn[ixf++]=argv[i];
        }

        // get the playlist from file
        for (i=0;i<npl;i++) {
            printf("-- Adding MP3 files from %s ...\n",pln[i]);
            if((fp=fopen(pln[i],"r"))==NULL) {perror("fopen"); exit(1);}
            for (;;) {
                fgets(buffer,1024,fp);
                if (feof(fp)) break;
                n=strlen(buffer);
                p=malloc(n);
                strncpy(p,buffer,n-1);
                pfn[ixf++]=p;
            }
            fclose(fp);
        }

        // Scan directory trees
        for (i=0;i<nd;i++) {
            printf("-- Scanning %s tree...\n",pdir[i]);
            dir_add(pdir[i],pfn,&ixf);
        }

        nf=ixf;
        printf("-- Total %d files\n",nf);
        if (!nf) exit(0);

        // Dump playlist
        if (df) {
            printf("-- Playlist dump to: %s\n",df);
            if ((fp=fopen(df,"w"))==NULL) exit(1);
            for (i=0;i<nf;i++) fprintf(fp,"%s\n",pfn[i]);
            fclose(fp);
        }

        // Random play stuff
        if (random) {
            printf("-- Random play\n"); fflush(stdout);
            srand(time(NULL));
            for (i=0;i<nf;i++) {
                j=rand()%nf;
                p=pfn[i]; pfn[i]=pfn[j]; pfn[j]=p;
            }
        }

        // stdin settings
        setnocanon();
        // stdin in non-block mode
        i=fcntl(0,F_GETFL);
        fcntl(0,F_SETFL,i|O_NONBLOCK);

        // Use root privileges (if possible) to
        // improve playing
        if (setegid (egid)<0) exit(1);
        if (seteuid (euid)<0) exit(1);

        // Starting reading trhead...
        rdfn=pfn[0];
        pthread_create(&pth, NULL, thread_read_file , NULL);
        while (rdfn) usleep(20000);

        // Try to set a high priority for this thread
        setpriority(PRIO_PROCESS,0,-20);

        for (ixf=0;ixf<nf;ixf++) {
            while (rdfn!=NULL) usleep(20000);
            p=bbuf; n=blen;
            if (ixf<nf-1) { rdfn=pfn[ixf+1];} // prefetch next file
            if (n>0) {
                do {
                    printf("\r-- Playing %s\n",pfn[ixf]);
                    i=play_block(p,n);
                } while (i=='r' || (i=='p' && ixf==0));
                munlock (p,n); // Allow swapping
                free(p);
                if (i=='q') break;
                if (i=='p') {
                    while (rdfn!=NULL) usleep(20000);
                    if (ixf<nf-1) free(bbuf); // discard prefetched file
                    ixf-=2;
                    rdfn=pfn[ixf+1]; // Read new file
                }
            }
        }
        setcanon();
        printf("\n");
}

